import struct
from math import sqrt,floor
import os.path
import string

FEATURE_FLAG_LEVEL = 0x00000001
FEATURE_FLAG_TOOLS = 0x00000002
FEATURE_FLAG_SOLUTION = 0x00000004

FeatureFlagNames = [
    "invalid",
    "NormalLevel",
    "Tools",
    "Solution",
]


LINKTYPE_ROPE = 0
LINKTYPE_CLOTH = 1
LINKTYPE_BAR = 2
LINKTYPE_PLATE = 3
LINKTYPE_ELASTIC = 4
LINKTYPE_RUBBER = 5
LINKTYPE_ROCKET = 6
LINKTYPE_MAX = 7


LinkNames = [
    "rope",
    "cloth",
    "bar",
    "plate",
    "elastic",
    "rubber",
    "rocket",
    "MAX"
]

POINT_TYPE_ANCHOR = 0
POINT_TYPE_MOBILE = 1
POINT_TYPE_USER_PLACED = 2
POINT_TYPE_MAX = 3

PointNames = [
    "Anchor",
    "Mobile",
    "User",
    "MAX"
]

LINK_FLAGS_USER = 1
LinkFlagUserNames = [
    "Editor",
    "User"
    ]

def castFloatToInt(f):
    (i,) = struct.unpack("I",struct.pack("f",f))
    return i


class DilloPoint:
    def __init__(self,level,x=0.0,y=0.0,type=POINT_TYPE_MOBILE):
        self.level = level
        self.links=[]
        self.index = -1
        self.x = x
        self.y = y
        self.type = type
        

    def isAnchor(self):
        return self.type == POINT_TYPE_ANCHOR

    def isMobile(self):
        return self.type == POINT_TYPE_MOBILE

    def isUser(self):
        return self.type == POINT_TYPE_USER_PLACED

    def isEditor(self):
        return not self.isUser()

    def calculateLevelChecksum(self):
        checksum = 0
        if self.isEditor():
            checksum ^= self.index
            checksum ^= self.type
            checksum ^= castFloatToInt(self.x)
            checksum ^= castFloatToInt(self.y)
        return checksum    

    def readData(self,file):
        pointData = file.read(4*4)
        (self.index,self.type,self.x,self.y)=struct.unpack("IIff",pointData)

    def writeData(self,file):
        file.write(struct.pack("IIff",self.index,self.type,self.x,self.y))

    def dump(self):
        print "\t\t%d %s (%f,%f)"%(self.index,PointNames[self.type],self.x,self.y)

class DilloLink:
    def __init__(self,level,fromIndex=-1,toIndex=-1,type=LINKTYPE_BAR):
        self.level = level
        self.group = None
        self.flags = 0
        self.type = type
        self.fromIndex = fromIndex
        if self.fromIndex!=-1:
            self.fromPoint = level.points[self.fromIndex]
        self.toIndex = toIndex
        if self.toIndex!=-1:
            self.toPoint = level.points[self.toIndex]
        self.hasTension = 0
        self.hasTimer = 0
        

    def isUser(self):
        return (self.flags & LINK_FLAGS_USER) != 0

    def isEditor(self):
        return not self.isUser()

    def costRope(self,length):
        baseCost=6+floor((length-0.8)/0.4)*2
        if self.hasTension:
            baseCost+=100
        return baseCost
    
    def costCloth(self,length):
        baseCost=12+floor((length-0.6)/0.3)*4
        if self.hasTension:
            baseCost+=100
        return baseCost
    
    def costBar(self,length):
        baseCost = 10
        if self.hasTension:
            baseCost+=100
        return baseCost
    
    def costPlate(self,length):
        baseCost = 20
        if self.hasTension:
            baseCost+=100
        return baseCost
    
    def costElastic(self,length):
        baseCost =90+floor((length-0.6)/0.3)*30
        if self.hasTension:
            baseCost+=100
        return baseCost
    
    def costRubber(self,length):
        baseCost = 100
        if self.hasTension:
            baseCost+=100
        return baseCost
    
    def costRocket(self,length):
        baseCost = 1000
        if self.hasTension:
            baseCost+=100
        return baseCost
    


    def cost(self):
        costFuncs = [
            DilloLink.costRope,
            DilloLink.costCloth,
            DilloLink.costBar,
            DilloLink.costPlate,
            DilloLink.costElastic,
            DilloLink.costRubber,
            DilloLink.costRocket
            ]
        return costFuncs[self.type](self,self.length())

    def calculateLevelChecksum(self):
        checksum = 0
        if self.isEditor():
            checksum ^= self.index
            checksum ^= self.flags
            checksum ^= self.type
            checksum ^= self.fromIndex
            checksum ^= self.toIndex
            checksum ^= self.hasTension
            if self.hasTension:
                checksum ^= castFloatToInt(self.tension)
            checksum ^= self.hasTimer
            if self.hasTimer:
                checksum ^=castFloatToInt(self.timer)
        return checksum    


    def readData(self,file):
        linkSize= 20
        linkData = file.read(linkSize)
        unpackedLink=struct.unpack("IIIII",linkData)
        (self.index,self.flags,self.type,self.fromIndex,self.toIndex)=unpackedLink
        self.fromPoint = self.level.points[self.fromIndex]
        self.toPoint = self.level.points[self.toIndex]
        self.tension = 0.0
        (self.hasTension,)=struct.unpack("B",file.read(1))
        if self.hasTension!=0:
            tensionData = file.read(4)
            (self.tension,) = struct.unpack("f",tensionData)
        self.timer = 0.0        
        (self.hasTimer,)=struct.unpack("B",file.read(1))
        if self.hasTimer!=0:
            timerData = file.read(4)
            (self.timer,) = struct.unpack("f",timerData)

    def writeData(self,file):
        file.write(struct.pack("IIIII",self.index,self.flags,self.type,self.fromIndex,self.toIndex))
        
        file.write(struct.pack("B",self.hasTension))
        if self.hasTension!=0:
            file.write(sruct.pack("f",self.tension))
            
        file.write(struct.pack("B",self.hasTimer))
        if self.hasTimer!=0:
            file.write(sruct.pack("f",self.timer))

    def length(self):
        dx = self.fromPoint.x - self.toPoint.x
        dy = self.fromPoint.y - self.toPoint.y
        return sqrt(dx*dx+dy*dy)
        
    def dump(self):
        if self.hasTimer:
            timerStr = " timer(%f(0x%08X))"%(self.timer,self.timer)
        else:
            timerStr = ""
        if self.hasTension:
            tensionStr = " tension(%f(0x%08X))"%(self.tension,self.tension)
        else:
            tensionStr = ""
        print "\t\t%d 0x%08X,%s,%d->%d (%f,%f)->(%f,%f)%s%s"%(self.index,self.flags,LinkNames[self.type],self.fromIndex,self.toIndex,self.fromPoint.x,self.fromPoint.y,self.toPoint.x,self.toPoint.y,timerStr,tensionStr)

class DilloLinkGroup:
    def __init__(self,level):
        self.level = level
        self.links = []
        self.linkIndices = []

    def calculateLevelChecksum(self):
        checksum = 0
        checksum ^= len(self.linkIndices)-1  # The -1 is to account for the funky way that PeterS stores groups in the level file
        for linkIndex in self.linkIndices:
            checksum ^= linkIndex
        return checksum    


    def addLinkIndex(self,i):
        
        self.linkIndices.append(i)
        link = self.level.links[i]
        self.links.append(link)
        link.group = self
        

    def readData(self,file):
        (startLinkIndex,) = struct.unpack("I",file.read(4))
        self.addLinkIndex(startLinkIndex)
        (count,) = struct.unpack("I",file.read(4))
        for i in range(0,count):
            (index,) = struct.unpack("I",file.read(4))
            self.addLinkIndex(index)

    def writeData(self,file):
        assert(len(self.linkIndices)>=2)
        file.write(struct.pack("I",self.linkIndices[0]))
        file.write(struct.pack("I",len(self.linkIndices)-1))
        for index in self.linkIndices[1:]:
            file.write(struct.pack("I",index))

        
    def dump(self):
        print "\t\t[%s]"%string.join(map(lambda i:"%d"%(i),self.linkIndices),",")

class DilloDillo:
    def __init__(self,level,dilloX=0.0,dilloY=0.0,dilloSize=0.5,portalX= 0.0,portalY=0.0,portalSize=1.0):
        self.level = level
        self.dilloX = dilloX
        self.dilloY = dilloY
        self.dilloSize = dilloSize
        self.portalX = portalX
        self.portalY = portalY
        self.portalSize = portalSize

    def calculateLevelChecksum(self):
        checksum = 0
        checksum ^=castFloatToInt(self.dilloX)
        checksum ^=castFloatToInt(self.dilloY)
        checksum ^=castFloatToInt(self.dilloSize)
        checksum ^=castFloatToInt(self.portalX)
        checksum ^=castFloatToInt(self.portalY)
        checksum ^=castFloatToInt(self.portalSize)
        return checksum    

    def readData(self,file):
        dilloDataSize= 12
        dilloData = file.read(dilloDataSize)
        unpackedDillo=struct.unpack("fff",dilloData)
        (self.dilloX,self.dilloY,self.dilloSize)=unpackedDillo
        portalDataSize= 12
        portalData = file.read(portalDataSize)
        unpackedPortal=struct.unpack("fff",portalData)
        (self.portalX,self.portalY,self.portalSize)=unpackedPortal

    def writeData(self,file):
        file.write(struct.pack("fff",self.dilloX,self.dilloY,self.dilloSize))
        file.write(struct.pack("fff",self.portalX,self.portalY,self.portalSize))
        
    def dump(self):
        print "\t\t(%f,%f)[%f] -> (%f,%f)[%f]"%(self.dilloX,self.dilloY,self.dilloSize,self.portalX,self.portalY,self.portalSize)
        

class DilloLevel:
    def __init__(self):
        self.points = []
        self.links = []
        self.groups = []
        self.dillos = []
        self.features = FEATURE_FLAG_LEVEL
        self.magic ="****"
        self.levelName = ""
        self.solutionName = ""
        self.background=0
        self.leftBound = -10.0
        self.topBound = 10.0
        self.rightBound = 10.0
        self.bottomBound = -10.0
        self.budget = 0
        self.cameraData = (0.688259,-4.210526,0.000000,0.000000,1.000000,0.000000,3.734861,0.000000,0.200000)
        self.tutorialText=""
        self.levelChecksum=0
        self.slnChecksum = (0,0)

        self.dilloOffset=-1
        self.pointOffset=-1
        self.linkOffset=-1
        self.cameraOffset=-1
        self.groupOffset = -1

    def isSolution(self):
        return (self.features&FEATURE_FLAG_SOLUTION)!=0
        
    def isTutorial(self):
        return (self.features&FEATURE_FLAG_LEVEL)==0
            
    def hasToolData(self):
        return (self.features&FEATURE_FLAG_TOOLS)!=0

    def addPoint(self,x,y,type):
        p = DilloPoint(self,x,y,type)
        p.index = len(self.points)
        self.points.append(p)
        return p

    def addLink(self,p1,p2,type):
        l = DilloLink(self,p1.index,p2.index,type)
        l.index = len(self.links)
        self.links.append(l)
        return l

    def addDillo(self,dilloX,dilloY,dilloSize,portalX,portalY,portalSize):
        d = DilloDillo(self,dilloX,dilloY,dilloSize,portalX,portalY,portalSize)
        self.dillos.append(d)
        return d

    def calculateLevelChecksum(self):
        checksum = 0x0000000C #I don't know why this doesn;t start at 0
        checksum ^= self.features
        checksum ^= self.background
        checksum ^= castFloatToInt(self.leftBound)
        checksum ^= castFloatToInt(self.rightBound)
        checksum ^= castFloatToInt(self.bottomBound)
        checksum ^= castFloatToInt(self.topBound)
        checksum ^= self.budget

        checksum ^= len(self.points)
        for point in self.points:
            checksum ^= point.calculateLevelChecksum()

        checksum ^= len(self.links)
        for link in self.links:
            checksum ^= link.calculateLevelChecksum()
        
        checksum ^= len(self.groups)
        for group in self.groups:
            checksum ^= group.calculateLevelChecksum()

        checksum ^= len(self.dillos)
        for dillo in self.dillos:
            checksum ^= dillo.calculateLevelChecksum()

        #checksumCamera
        for c in self.cameraData:
            checksum ^= castFloatToInt(c)

        
        if (self.features&FEATURE_FLAG_TOOLS)!=0:
            for tool in self.tools:
                checksum ^= tool

        if (self.features&FEATURE_FLAG_LEVEL)==0: #if it is a tutorial
            for c in self.tutorialText:
                checksum ^= ord(c)

        return checksum                    
            

    def load(self,filename):
        self.filename = os.path.split(filename)[1] # remove the directory
        levelname = os.path.splitext(self.filename)[0] # remove the ".lvl"
        filenameBits = string.split(levelname,".",1)
        self.levelName = filenameBits[0]
        if len(filenameBits)>1:
            self.solutionName = filenameBits[1]
        else:
            self.solutionName = ""
            
        
        f = open(filename,"rb")

        f.seek(0,2)
        size = f.tell()
        f.seek(0,0)

        self.readHeader(f)
        self.readPoints(f)
        self.readLinks(f)
        self.readGroups(f)
        self.readDillos(f)
        self.readCamera(f)
        self.readToolFlags(f)
        self.readTutorialText(f)
        self.readChecksum(f)

        readDataSize= f.tell()
        f.seek(0,2)
        assert(readDataSize==f.tell())
        
        
        f.close()

    def readHeader(self,file):
        headerFmt="4sIIffffi"
        headerSize = 8*4
        data = file.read(headerSize)
        unpackedData = struct.unpack(headerFmt,data)
        (self.magic,self.features,self.background,self.leftBound,self.rightBound,self.bottomBound,self.topBound,self.budget) = unpackedData

    def readPoints(self,file):
        self.pointOffset = file.tell()
        countData=file.read(4)
        (numPoints,)=struct.unpack("I",countData)

        for i in range(0,numPoints):
            point = DilloPoint(self)
            point.readData(file)
            self.points.append(point)

    def readLinks(self,file):
        self.linkOffset = file.tell()
        countData = file.read(4)
        (numLinks,)=struct.unpack("I",countData)
        for i in range(0,numLinks):
            link = DilloLink(self)
            link.readData(file)
            self.links.append(link)

    def readGroups(self,file):
        self.groupOffset = file.tell()
        countData = file.read(4)
        (numGroups,)=struct.unpack("I",countData)
        for i in range(0,numGroups):
            group = DilloLinkGroup(self)
            group.readData(file)
            self.groups.append(group)

    def readDillos(self,file):
        self.dilloOffset = file.tell()
        dilloCountData = file.read(4)
        (numDillos,)=struct.unpack("I",dilloCountData)
        for i in range(0,numDillos):
            dillo = DilloDillo(self)
            dillo.readData(file)
            self.dillos.append(dillo)


    def readCamera(self,file):
        cameraSize = 36
        self.cameraData = struct.unpack("9f",file.read(cameraSize))

    def readToolFlags(self,file):
        if (self.features&FEATURE_FLAG_TOOLS)!=0:
            toolData = file.read(10)
            self.tools = struct.unpack("10B",toolData)
        else:
            self.tools = []
            
    def readTutorialText(self,file):
        self.tutorialText = ""
        if (self.features&FEATURE_FLAG_LEVEL)==0: #if it is a tutorial
            c=file.read(1)
            while ord(c)!=0:
                self.tutorialText+=c
                c=file.read(1)
    
        
    def readChecksum(self,file):
        self.checksumOffset = file.tell()
        #print "0x%08X"%(self.checksumOffset)
        levelCheckSumData = file.read(4)
        (self.levelChecksum,)=struct.unpack("I",levelCheckSumData)
        if (self.features&FEATURE_FLAG_SOLUTION)!=0:
            slnCheckSumData = file.read(8)
            self.slnChecksum=struct.unpack("II",slnCheckSumData)
        else:
            self.slnChecksum = (0,0)            

    def save(self,filename):
        self.filename = os.path.split(filename)[1] # remove the directory
        levelname = os.path.splitext(self.filename)[0] # remove the ".lvl"
        filenameBits = string.split(levelname,".",1)
        self.levelName = filenameBits[0]
        if len(filenameBits)>1:
            self.solutionName = filenameBits[1]
        else:
            self.solutionName = ""
            
        
        f = open(filename,"wb")

        self.writeHeader(f)
        self.writePoints(f)
        self.writeLinks(f)
        self.writeGroups(f)
        self.writeDillos(f)
        self.writeCamera(f)
        self.writeToolFlags(f)
        self.writeTutorialText(f)
        self.writeChecksum(f)
        f.close()

    def writeHeader(self,file):
        headerFmt="4sIIffffi"
        self.magic = "ArmR"
        file.write(struct.pack(headerFmt,self.magic,self.features,self.background,self.leftBound,self.rightBound,self.bottomBound,self.topBound,self.budget))
        
    def writePoints(self,file):
        self.pointOffset = file.tell()
        file.write(struct.pack("I",len(self.points)))
        for point in self.points:
            point.writeData(file)
        
    def writeLinks(self,file):
        self.linkOffset = file.tell()
        file.write(struct.pack("I",len(self.links)))
        for link in self.links:
            link.writeData(file)

    def writeGroups(self,file):
        self.groupOffset = file.tell()
        file.write(struct.pack("I",len(self.groups)))
        for group in self.groups:
            group.writeData(file)

    def writeDillos(self,file):
        self.dilloOffset = file.tell()
        file.write(struct.pack("I",len(self.dillos)))
        for dillo in self.dillos:
            dillo.writeData(file)
            
    def writeCamera(self,file):
        assert(len(self.cameraData)==9)
        file.write(struct.pack("9f",*self.cameraData))
            
    def writeToolFlags(self,file):
        if (self.features&FEATURE_FLAG_TOOLS)!=0:
            file.write(struct.pack("10B",self.tools))
        else:
            pass

    def writeTutorialText(self,file):
        if (self.features&FEATURE_FLAG_LEVEL)==0: #if it is a tutorial
            file.write(self.tutorialText)
            file.write(struct.pack("B",0)) #Null terminator
            
    def writeChecksum(self,file):
        self.checksumOffset = file.tell()
        levelCheckSum = self.calculateLevelChecksum()
        file.write(struct.pack("I",levelCheckSum))
        if (self.features&FEATURE_FLAG_SOLUTION)!=0:
            file.write(struct.pack("II",self.slnChecksum))
        else:
            pass
               
        

    def dump(self):
        if self.solutionName!="":
            print "%s (%s)"%(self.levelName,self.solutionName)
        else:
            print "%s"%(self.levelName)
        print "\tmagic: %s"%(self.magic)
        print "\tfeatures: 0x%08X"%(self.features)
        print "\tbackground: %d"%(self.background)
        print "\tbounds(l,r,b,t): (%f,%f,%f,%f)"%(self.leftBound,self.rightBound,self.bottomBound,self.topBound)
        print "\tbudget: $%d"%(self.budget)
        print "\tcameraData: [%s]"%(string.join(map(lambda i:"%f"%i,list(self.cameraData)),","))
        print "\ttext:\"%s\""%(self.tutorialText)
        print "\tchecksum: 0x%08X (0x%08X,0x%08X)"%((self.levelChecksum,)+self.slnChecksum)
        print "\t%d Dillos (0x%08X): "%(len(self.dillos),self.dilloOffset)
        for dillo in self.dillos:
            dillo.dump()
        print "\t%d Points (0x%08X): "%(len(self.points),self.pointOffset)
        for point in self.points:
            point.dump()
        print "\t%d Links (0x%08X):"%(len(self.links),self.linkOffset)
        for link in self.links:
            link.dump()
        print "\t%d Groups (0x%08X):"%(len(self.groups),self.groupOffset)
        for group in self.groups:
            group.dump()


